package com.github.polpetta.mezzotre.telegram.callbackquery;

import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import io.vavr.control.Try;
import java.nio.charset.StandardCharsets;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.util.StringBuilderWriter;
import org.slf4j.Logger;

/**
 * ChangeLanguage event is related to support locale change for a particular chat
 *
 * @author Davide Polonio
 * @since 1.0
 */
@Singleton
public class ChangeLanguage implements Processor {

  public static final String EVENT_NAME = "changeLanguage";
  private final Executor threadPool;
  private final LocalizedMessageFactory localizedMessageFactory;
  private final Logger log;

  /**
   * Additional fields that are related to {@code changeLanguage} event
   *
   * @author Davide Polonio
   * @since 1.0
   */
  public enum Field {
    NewLanguage("newLanguage");

    private final String name;

    Field(String name) {
      this.name = name;
    }

    public String getName() {
      return name;
    }
  }

  /**
   * Possible values for the additional {@link Field} of {@code changeLanguage} event
   *
   * @author Davide Polonio
   * @since 1.0
   */
  public enum Language {
    English("en-US"),
    Italian("it-IT");

    private final String locale;

    Language(String locale) {
      this.locale = locale;
    }

    public String getLocale() {
      return locale;
    }
  }

  @Inject
  public ChangeLanguage(
      @Named("eventThreadPool") Executor threadPool,
      LocalizedMessageFactory localizedMessageFactory,
      Logger log) {
    this.threadPool = threadPool;
    this.localizedMessageFactory = localizedMessageFactory;
    this.log = log;
  }

  @Override
  public String getEventName() {
    return EVENT_NAME;
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
      CallbackQueryContext callbackQueryContext, Update update) {
    return CompletableFuture.supplyAsync(
            () ->
                Optional.of(callbackQueryContext.getFields().getTelegramChatId())
                    .filter(
                        chatId ->
                            chatId.longValue() != 0L
                                && !chatId.isNaN()
                                && !chatId.isInfinite()
                                && chatId != Double.MIN_VALUE)
                    .flatMap(chatId -> new QTgChat().id.eq(chatId.longValue()).findOneOrEmpty())
                    .map(
                        tgChat -> {
                          tgChat.setLocale(
                              (String)
                                  callbackQueryContext
                                      .getFields()
                                      .getAdditionalProperties()
                                      .getOrDefault(
                                          Field.NewLanguage.getName(),
                                          Language.English.getLocale()));
                          tgChat.save();
                          return tgChat;
                        })
                    .orElseThrow(
                        () ->
                            new NoSuchElementException(
                                "Unable to find telegram chat "
                                    + Double.valueOf(
                                            callbackQueryContext.getFields().getTelegramChatId())
                                        .longValue()
                                    + " in the database")),
            threadPool)
        .thenApplyAsync(
            // If we are here then we're sure there is at least a chat associated with this callback
            tgChat -> {
              final String message =
                  Try.of(
                          () -> {
                            final Locale locale = Locale.forLanguageTag(tgChat.getLocale());
                            final ToolManager toolManager = localizedMessageFactory.create(locale);
                            final VelocityContext velocityContext =
                                new VelocityContext(toolManager.createContext());

                            final StringBuilder content = new StringBuilder();
                            final StringBuilderWriter stringBuilderWriter =
                                new StringBuilderWriter(content);

                            toolManager
                                .getVelocityEngine()
                                .mergeTemplate(
                                    "/template/callbackQuery/changeLanguage.vm",
                                    StandardCharsets.UTF_8.name(),
                                    velocityContext,
                                    stringBuilderWriter);

                            stringBuilderWriter.close();
                            return content.toString();
                          })
                      .get();

              log.trace("ChangeLanguage event - message to send back: " + message);

              final String callBackGroupToDelete = callbackQueryContext.getEntryGroup();
              final int delete =
                  new QCallbackQueryContext().entryGroup.eq(callBackGroupToDelete).delete();
              log.trace(
                  "Deleted "
                      + delete
                      + " entries regarding callback group "
                      + callBackGroupToDelete);

              return Optional.of(
                  new SendMessage(tgChat.getId(), message).parseMode(ParseMode.Markdown));
            },
            threadPool);
  }
}
