package com.github.polpetta.mezzotre.telegram.command;

import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.telegram.callbackquery.ChangeLanguage;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.github.polpetta.types.json.ChatContext;
import com.google.inject.Singleton;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import io.vavr.control.Try;
import java.nio.charset.StandardCharsets;
import java.util.Locale;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.util.StringBuilderWriter;
import org.slf4j.Logger;

/**
 * This {@link Processor} has the goal to greet a user that typed {@code /start} to the bot.
 *
 * @author Davide Polonio
 * @since 1.0
 */
@Singleton
public class Start implements Processor {

  private final Executor threadPool;
  private final Logger log;
  private final UUIDGenerator uuidGenerator;
  private final Clock clock;
  private final LocalizedMessageFactory localizedMessageFactory;

  @Inject
  public Start(
      LocalizedMessageFactory localizedMessageFactory,
      @Named("eventThreadPool") Executor threadPool,
      Logger log,
      UUIDGenerator uuidGenerator,
      Clock clock) {
    this.localizedMessageFactory = localizedMessageFactory;
    this.threadPool = threadPool;
    this.log = log;
    this.uuidGenerator = uuidGenerator;
    this.clock = clock;
  }

  @Override
  public String getTriggerKeyword() {
    return "/start";
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(TgChat chat, Update update) {
    return CompletableFuture.supplyAsync(
        () -> {
          /*
          Steps:
            1 - Load the template and generate the message with the defined locale
            2 - Update the chat context
            3 - Reply to Telegram
           */
          final String message =
              Try.of(
                      () -> {
                        final Locale locale = Locale.forLanguageTag(chat.getLocale());
                        final ToolManager toolManager = localizedMessageFactory.create(locale);
                        final VelocityContext context =
                            new VelocityContext(toolManager.createContext());
                        context.put("firstName", update.message().chat().firstName());
                        context.put("programName", "_Mezzotre_");

                        final StringBuilder content = new StringBuilder();
                        final StringBuilderWriter stringBuilderWriter =
                            new StringBuilderWriter(content);

                        toolManager
                            .getVelocityEngine()
                            .mergeTemplate(
                                "template/command/start.0.vm",
                                StandardCharsets.UTF_8.name(),
                                context,
                                stringBuilderWriter);

                        stringBuilderWriter.close();
                        return content.toString();
                      })
                  .get();
          log.trace("Start command - message to send back: " + message);

          final ChatContext chatContext = chat.getChatContext();
          chatContext.setStage(getTriggerKeyword());
          chatContext.setStep(0);
          chatContext.setPreviousMessageUnixTimestampInSeconds(clock.now());
          chat.setChatContext(chatContext);
          chat.save();

          // To get the messageId we should send the message first, then save it in the database!
          final String groupId = uuidGenerator.generateAsString();
          final CallbackQueryContext switchToEnglish =
              new CallbackQueryContext(
                  uuidGenerator.generateAsString(),
                  groupId,
                  new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                      .withEvent(ChangeLanguage.EVENT_NAME)
                      .withTelegramChatId(update.message().chat().id())
                      .withAdditionalProperty(
                          ChangeLanguage.Field.NewLanguage.getName(),
                          ChangeLanguage.Language.English)
                      .build());

          final CallbackQueryContext switchToItalian =
              new CallbackQueryContext(
                  uuidGenerator.generateAsString(),
                  groupId,
                  new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                      .withEvent(ChangeLanguage.EVENT_NAME)
                      .withTelegramChatId(update.message().chat().id())
                      .withAdditionalProperty(
                          ChangeLanguage.Field.NewLanguage.getName(),
                          ChangeLanguage.Language.Italian)
                      .build());

          final SendMessage messageToSend =
              new SendMessage(chat.getId(), message)
                  .parseMode(ParseMode.Markdown)
                  .replyMarkup(
                      new InlineKeyboardMarkup(
                          new InlineKeyboardButton("English").callbackData(switchToEnglish.getId()),
                          new InlineKeyboardButton("Italian")
                              .callbackData(switchToItalian.getId())));

          switchToEnglish.save();
          switchToItalian.save();

          return Optional.of(messageToSend);
        },
        threadPool);
  }
}
