package com.github.polpetta.mezzotre.route;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.telegram.callbackquery.Dispatcher;
import com.github.polpetta.mezzotre.telegram.command.Router;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.pengrad.telegrambot.TelegramBot;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import io.jooby.Context;
import io.jooby.MediaType;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import org.junit.jupiter.api.*;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Testcontainers
class TelegramIntegrationTest {

  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private Telegram telegram;
  private TelegramBot fakeTelegramBot;
  private Router fakeRouter;
  private Dispatcher fakeDispatcher;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));

    fakeTelegramBot = mock(TelegramBot.class);
    fakeRouter = mock(Router.class);
    fakeDispatcher = mock(Dispatcher.class);

    telegram =
        new Telegram(
            LoggerFactory.getLogger(getClass()),
            new GsonBuilder().setPrettyPrinting().create(),
            Executors.newSingleThreadExecutor(),
            fakeRouter,
            fakeDispatcher);
  }

  @Test
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldReceiveAValidTelegramMessageUpdateAsResponse() throws Exception {

    final SendMessage expectedBaseRequest = new SendMessage(1111111, "Hello world");
    when(fakeRouter.process(any(TgChat.class), any(Update.class)))
        .thenReturn(CompletableFuture.completedFuture(Optional.of(expectedBaseRequest)));

    final Context fakeContext = mock(Context.class);
    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/start\"\n"
                + "}\n"
                + "}\n",
            Update.class);
    final CompletableFuture<String> gotResponseFuture =
        telegram.incomingUpdate(fakeContext, update);
    final String gotReply = gotResponseFuture.get();
    verify(fakeContext, times(1)).setResponseType(MediaType.JSON);
    verify(fakeRouter, times(1)).process(any(), any());
    verify(fakeDispatcher, times(0)).dispatch(any(), any());
    assertEquals(expectedBaseRequest.toWebhookResponse(), gotReply);
  }

  @Test
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldProcessAnIncomingCallbackQueryThatExistsInTheDb() throws Exception {
    final SendMessage expectedBaseRequest = new SendMessage(1111111, "Hello world");
    when(fakeDispatcher.dispatch(any(CallbackQueryContext.class), any(Update.class)))
        .thenReturn(CompletableFuture.completedFuture(Optional.of(expectedBaseRequest)));
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext(
            "41427473-0d81-40a8-af60-9517163615a4",
            "2ee7f5c6-93f0-4859-b902-af9476cf74ad",
            new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                .withTelegramChatId(666L)
                .withMessageId(42L)
                .withEvent("testEvent")
                .build());
    callbackQueryContext.save();

    final Context fakeContext = mock(Context.class);
    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"callback_query\":{\n"
                + "  \"id\": \"4382bfdwdsb323b2d9\",\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"type\": \"private\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"data\": \"41427473-0d81-40a8-af60-9517163615a4\",\n"
                + "  \"inline_message_id\": \"1234csdbsk4839\"\n"
                + "}\n"
                + "}",
            Update.class);
    final CompletableFuture<String> gotResponseFuture =
        telegram.incomingUpdate(fakeContext, update);
    final String gotReply = gotResponseFuture.get();
    verify(fakeContext, times(1)).setResponseType(MediaType.JSON);
    verify(fakeRouter, times(0)).process(any(), any());
    verify(fakeDispatcher, times(1)).dispatch(eq(callbackQueryContext), any());
    assertEquals(expectedBaseRequest.toWebhookResponse(), gotReply);
  }
}
