package com.github.polpetta.mezzotre.telegram.callbackquery;

import static org.junit.jupiter.api.Assertions.*;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;
import org.junit.jupiter.api.*;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class ChangeLanguageIntegrationTest {

  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private ChangeLanguage changeLanguage;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));

    changeLanguage =
        new ChangeLanguage(
            Executors.newSingleThreadExecutor(),
            new LocalizedMessageFactory(Loader.defaultVelocityEngine()),
            LoggerFactory.getLogger(ChangeLanguage.class));
  }

  private static Stream<Arguments> getTestLocales() {
    return Stream.of(
        Arguments.of(
            ChangeLanguage.Language.Italian,
            "_*Procede a bere una pozione al cui suo interno si trova uno strano liquido"
                + " multicolore*_\n"
                + "\n"
                + "Grazie! Ora che ho bevuto quest posizione modificata di Parlare con animali che"
                + " ho trovato ieri al negozio di pozioni magiche la \"Cristalleria Fermentatrice\""
                + " posso parlare con te nel linguaggio che preferisci!\n"
                + "\n"
                + "Puoi sempre cambiare le preferenze della tua lingua scrivendo /changeLanguage"
                + " nella chat.",
            "en-US"),
        Arguments.of(
            ChangeLanguage.Language.English,
            "_*Proceeds to drink a potion with a strange, multicolor liquid*_\n"
                + "\n"
                + "Thanks! Now that I drank this modified potion of Speak with animals that I've"
                + " found at the \"Crystal Fermentary\" magic potion shop yesterday I can speak"
                + " with you in the language you prefer!\n"
                + "\n"
                + "You can always change your language settings by typing /changeLanguage in the"
                + " chat.",
            "it-IT"));
  }

  @ParameterizedTest
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  @MethodSource("getTestLocales")
  void shouldProcessChangeLanguageToDesiredOne(
      ChangeLanguage.Language language, String expectedResult, String startingLocale)
      throws Exception {

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"callback_query\":{\n"
                + "  \"id\": \"4382bfdwdsb323b2d9\",\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"type\": \"private\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"data\": \"Data from button callback\",\n"
                + "  \"inline_message_id\": \"1234csdbsk4839\"\n"
                + "}\n"
                + "}",
            Update.class);

    final long tgChatId = 1111111L;
    final ChatContext chatContext = new ChatContext();
    final TgChat tgChat = new TgChat(tgChatId, chatContext, startingLocale);
    tgChat.save();

    final CallbackQueryMetadata callbackQueryMetadata =
        new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
            .withEvent("changeLanguage")
            .withTelegramChatId(tgChatId)
            .withAdditionalProperty(
                ChangeLanguage.Field.NewLanguage.getName(), language.getLocale())
            .build();
    final String entryGroupId = "2e67774a-e4e4-4369-a414-a7f8bfe74b80";
    final CallbackQueryContext changeLanguageCallbackQueryContext =
        new CallbackQueryContext(
            "c018108f-6612-4848-8fca-cf301460d4eb", entryGroupId, callbackQueryMetadata);
    changeLanguageCallbackQueryContext.save();

    final CompletableFuture<Optional<BaseRequest<?, ?>>> processFuture =
        changeLanguage.process(changeLanguageCallbackQueryContext, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = processFuture.get();
    final SendMessage gotMessage = (SendMessage) gotResponseOpt.get();
    assertEquals(expectedResult, gotMessage.getParameters().get("text"));

    final TgChat retrievedTgChat = new QTgChat().id.eq(tgChatId).findOne();
    assertNotNull(retrievedTgChat);
    assertEquals(language.getLocale(), retrievedTgChat.getLocale());

    assertEquals(0, new QCallbackQueryContext().entryGroup.eq(entryGroupId).findCount());
  }
}
