package com.github.polpetta.mezzotre.telegram.command;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;

@Execution(ExecutionMode.CONCURRENT)
class RouterTest {

  private static Processor dummyEmptyExampleProcessor;
  private static Processor anotherKeyWithResultProcessor;
  private static Gson gson;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();

    dummyEmptyExampleProcessor = mock(Processor.class);
    when(dummyEmptyExampleProcessor.getTriggerKeyword()).thenReturn("/example");
    when(dummyEmptyExampleProcessor.process(any(), any()))
        .thenReturn(CompletableFuture.completedFuture(Optional.empty()));

    anotherKeyWithResultProcessor = mock(Processor.class);
    when(anotherKeyWithResultProcessor.getTriggerKeyword()).thenReturn("/anotherExample");
    when(anotherKeyWithResultProcessor.process(any(), any()))
        .thenReturn(
            CompletableFuture.completedFuture(Optional.of(new SendMessage(1234L, "hello world"))));
  }

  @Test
  void shouldMessageExampleMessageAndGetEmptyOptional() throws Exception {
    final Router router =
        new Router(Set.of(dummyEmptyExampleProcessor), Executors.newSingleThreadExecutor());
    final TgChat fakeChat = mock(TgChat.class);
    when(fakeChat.getChatContext()).thenReturn(new ChatContext());
    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/example\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotExecution =
        router.process(fakeChat, update);
    assertDoesNotThrow(() -> gotExecution.get());
    final Optional<BaseRequest<?, ?>> gotRequest = gotExecution.get();
    assertTrue(gotRequest.isEmpty());
  }

  @Test
  void shouldSelectRightExecutorAndReturnResult() throws Exception {
    final Router router =
        new Router(
            Set.of(dummyEmptyExampleProcessor, anotherKeyWithResultProcessor),
            Executors.newSingleThreadExecutor());
    final TgChat fakeChat = mock(TgChat.class);
    when(fakeChat.getChatContext()).thenReturn(new ChatContext());
    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/anotherExample\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotExecution =
        router.process(fakeChat, update);
    assertDoesNotThrow(() -> gotExecution.get());
    final Optional<BaseRequest<?, ?>> gotRequestOpt = gotExecution.get();
    assertDoesNotThrow(gotRequestOpt::get);
    final BaseRequest<?, ?> gotMessage = gotRequestOpt.get();
    assertInstanceOf(SendMessage.class, gotMessage);
    final String message = (String) gotMessage.getParameters().get("text");
    assertEquals("hello world", message);
    assertEquals(1234L, (Long) gotMessage.getParameters().get("chat_id"));
  }
}
